import vf3py
import networkx as nx

# Hardcoded sanity check test case.
def main():
    g1 = nx.Graph()
    g1.add_edges_from([(0, 1), (1, 2), (2, 3), (3, 0), (0, 2), (1, 3)])

    g2 = nx.Graph()
    g2.add_edges_from([('A','B'), ('B', 'C')])

    assert not vf3py.has_subgraph(g2, g1)
    assert vf3py.has_monomorphic_subgraph(g2, g1)

    sols = vf3py.get_subgraph_monomorphisms(g2, g1)
    # sols = vf3py.main_vf3_caller(g2, g1, induced=False)
    assert len(sols) == 24

    print("Monomorphism test passed")

if __name__ == "__main__":
    main()




import sys
import vf3py
import networkx as nx
from networkx.algorithms import isomorphism

TEST_GRAPHS = {
    'gnp_random_graph': lambda: nx.gnp_random_graph(5, 0.3),
    'cycle_graph': lambda: nx.cycle_graph(4),
    'star_graph': lambda: nx.star_graph(5),
    'complete_graph': lambda: nx.complete_graph(4),
    'barbell_graph': lambda: nx.barbell_graph(5, 0),
    'path_graph': lambda: nx.path_graph(4),
    'random_regular_graph': lambda: nx.random_regular_graph(3, 4),
    'watts_strogatz_graph': lambda: nx.watts_strogatz_graph(4, 3, 0.1),
    # 'erdos_renyi_graph': lambda: nx.erdos_renyi_graph(4, 0.2),
}

CRASH_GRAPHS = {
    'scale_free_graph': {
        'get_graph': lambda: nx.scale_free_graph(4),
        'exception': vf3py.ApplicabilityScopeError,
    }
}

def networkx_isomorphisms(G):
    GM = isomorphism.GraphMatcher(G, G)
    isomorphisms = set()
    atoms = list(G.nodes)
    for isom in GM.isomorphisms_iter():
        isomorphisms.add(tuple(isom[i] for i in atoms))
    return isomorphisms


def main(verbose=False):
    for graph_name, get_graph in TEST_GRAPHS.items():
        graph = get_graph()
        print(f"Processing test case '{graph_name}'")
        reference_answer = networkx_isomorphisms(graph)
        atoms = list(graph.nodes)
        if verbose:
            print([node for node in graph.nodes])
            print([edge for edge in graph.edges])
        if graph.number_of_edges() == 0:
            continue
        result = set(
            tuple(
                isom[i]
                for i in atoms
            )
            for isom in vf3py.get_automorphisms(graph, verbose=verbose)
        )
        assert result == reference_answer, \
            f"Failed for test graph '{graph_name}': correct={repr(reference_answer)},\ngot={repr(result)}"
    
    for graph_name, testcase_data in CRASH_GRAPHS.items():
        print(f"Processing test case '{graph_name}'")
        graph = testcase_data['get_graph']()
        expected_exception_type = testcase_data['exception']

        error = None
        try:
            vf3py.get_automorphisms(graph, verbose=verbose)
        except Exception as e:
            error = type(e)
        
        assert error == expected_exception_type, \
            f"Expected to get an error '{expected_exception_type}' but got '{error}'"
    
    print('Basic graph tests passed')


if __name__ == "__main__":
    VERBOSE = '-v' in sys.argv
    main(verbose=VERBOSE)



__version__ = '0.0.10'
import platform
import string
import multiprocessing
from typing import (
    Union,
    Callable,
    Tuple,
    List,
    Dict,
    Optional,
    Literal,
    Iterable,
    Set,
    Mapping,
)
import networkx as nx

assert platform.system() != "Windows", "Only Linux platform is supported in newer releases"
# if platform.system() == "Windows":
#     mypath = os.path.dirname(os.path.realpath(__file__))
#     if mypath not in sys.path:
#         sys.path.insert(0, mypath)
#     os.add_dll_directory(mypath)

from .cpppart import cpppart

VF3_VARIANTS = ('B', 'P', 'L')
VF3_VARIANTS_TYPE = Literal[VF3_VARIANTS]
MAX_THREADS = multiprocessing.cpu_count()

DOC_STRINGS = {
    'FIRST_GRAPH': "main_graph (nx.Graph | nx.DiGraph): First graph.",
    'SECOND_GRAPH': "ref_graph (nx.Graph | nx.DiGraph): Second graph.",
    'ANY_GRAPH': "graph (nx.Graph | nx.DiGraph): The graph of interest.",
    'LARGE_GRAPH': "graph (nx.Graph | nx.DiGraph): Searching for maps *into* this graph.",
    'SUBGRAPH': "subgraph (nx.Graph | nx.DiGraph): Searching for maps *of* this graph.",
    'NODE_MATCH': "node_match (Callable[[dict, dict], bool], optional): Nodes are allowed to be matched only if this function returns ``True`` for the nodes' attributes. Order of arguments: (1) dict of the source graph/subgraph, (2) target graph. Defaults to ``None``.",
    'EDGE_MATCH': "edge_match (Callable[[dict, dict], bool], optional): Edges are allowed to be matched only if this function returns ``True`` for the edges' attributes. Order of arguments: (1) dict of the source graph/subgraph, (2) target graph. Defaults to ``None``.",
    'ALL_SOLUTIONS': "all_solutions (bool, optional): Whether to generate all possible subgraph->graph matches (``True``) or only one (``False``). Defaults to ``True``.",
    'RETURN_INTEGERS': "return_integers (bool, optional): Whether to represent isomorphisms using integers (``True``) or the original labels of NetworkX graphs (``False``). Defaults to ``False``.",
    'VARIANT': f"variant (str, optional): The VF3 variant to be used. One of ``{VF3_VARIANTS}``. Defaults to ``'B'``.",
    'NUM_THREADS': "num_threads (int, optional): Number of threads to be used in parallel variant. Defaults to ``1``.",
    'VERBOSE': "verbose (bool, optional): Whether to print info on some intermediate steps. Defaults to ``False``.",
    'INDUCED': "induced (bool, optional): Whether to search for monomorphisms (``False``) or isomorphisms (``True``). Iso does not allow the target graph to have additional edges that are not present in the source graph. Defaults to ``True``.",
    'BIJECTIVE_FOR_NODES': "bijective_for_nodes (bool, optional): Whether to search for mono/isomorphisms between entire graphs (``True``) or allow to match with a subgraph of the target graph (``False``). Defaults to ``False``.",
}

GRAPH_TYPE = Union[nx.Graph, nx.DiGraph]

class NetworkxView(Set, Mapping):
    pass

def _keys_as_list(mask: str) -> list:
    return [
        t[1]
        for t in string.Formatter().parse(mask)
        if t[1] is not None
    ]


def _format_docstring(func: Callable) -> Callable:
    # Get the original docstring
    original_docstring = func.__doc__ if func.__doc__ else ""

    # Substitute placeholders with actual values
    formatted_docstring = original_docstring.format(**{
        key: DOC_STRINGS[key]
        for key in _keys_as_list(original_docstring)
    })

    # Set the new formatted docstring to the function
    func.__doc__ = formatted_docstring

    return func


VF3_CALLS = {
    # (variant, use_node_attrs, use_edge_attrs) => Appropriate function of the base pybind11 module
    ('B', False, False) : cpppart.calc_noattrs,
    ('B', True, False)  : cpppart.calc_nodeattr,
    ('B', False, True)  : cpppart.calc_edgeattr,
    ('B', True, True)   : cpppart.calc_bothattrs,
    ('L', False, False) : cpppart.calc_l_noattrs,
    ('L', True, False)  : cpppart.calc_l_nodeattr,
    ('L', False, True)  : cpppart.calc_l_edgeattr,
    ('L', True, True)   : cpppart.calc_l_bothattrs,
    ('P', False, False) : cpppart.calc_p_noattrs,
    ('P', True, False)  : cpppart.calc_p_nodeattr,
    ('P', False, True)  : cpppart.calc_p_edgeattr,
    ('P', True, True)   : cpppart.calc_p_bothattrs,
}


class ApplicabilityScopeError(Exception):
    """VF3Py has two limitations:

    * ``nx.MultiGraph`` and ``nx.MultiDiGraph`` are not supported.

    * Complex rules for matching nodes and edges. In particular, can not allow the same node (or edge) to be matched with multiple 'colors'.
    
    This exception can be thrown by any of these functions: ``are_isomorphic``, ``get_automorphisms``, ``get_exact_isomorphisms``, ``get_subgraph_isomorphisms``, ``has_subgraph``, ``main_vf3_caller``.
    """
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


def _ensure_graph_correctness(graph: GRAPH_TYPE) -> None:
    if isinstance(graph, nx.MultiGraph) or isinstance(graph, nx.MultiDiGraph):
        raise ApplicabilityScopeError("Cannot accept Multigraph type for isomorphism calculations")

    assert isinstance(graph, nx.Graph) or isinstance(graph, nx.DiGraph), \
        f"Cannot accept graph of type '{type(graph)}' (nx.Graph or nx.DiGraph is expected)"

    assert graph.number_of_nodes() > 0, \
        "Graph must contain non-zero number of nodes"

    if graph.number_of_edges() == 0:
        raise ApplicabilityScopeError("Cannot accept Graph with no edges")


def _process_graph(graph: GRAPH_TYPE) -> Tuple[Dict, Dict, bool]:
    assert isinstance(graph, nx.Graph) or isinstance(graph, nx.DiGraph), \
        f"Provided '{repr(graph)}' is not a NetworkX graph"

    directed = isinstance(graph, nx.DiGraph)

    int_relabel = [
        node
        for node in graph.nodes
    ]

    relabeling = {
        'to_int': {
            node: i
            for i, node in enumerate(int_relabel)
        },
        'from_int': int_relabel
    }

    return {
        'nodes': [i for i in range(graph.number_of_nodes())],
        'edges': [
            [relabeling['to_int'][vA], relabeling['to_int'][vB]]
            for vA, vB in graph.edges
        ],

        # Attr lists will be filled later
        'node_attrs': [],
        'edge_attrs': [],
    }, relabeling, directed


def _group_attrs(
        itemviewA: NetworkxView, # Nodes/edges of the two graphs
        itemviewB: NetworkxView,
        match_function,
        item_type: Literal['node', 'edge'],
        bijective=False,
    ) -> Tuple[Union[Dict, None], Union[Dict, None]]:

    assert match_function is not None

    dep_graph = nx.Graph()
    a_nodes = []
    i = 0
    for item_name in itemviewA:
        dep_graph.add_node(i, base='A', name=item_name)
        a_nodes.append(i)
        i += 1

    b_nodes = []
    for item_name in itemviewB:
        dep_graph.add_node(i, base='B', name=item_name)
        b_nodes.append(i)
        i += 1

    for a_node in a_nodes:
        for b_node in b_nodes:
            if match_function(itemviewA[dep_graph.nodes[a_node]['name']], itemviewB[dep_graph.nodes[b_node]['name']]):
                dep_graph.add_edge(a_node, b_node)

    a_attrs = {}
    b_attrs = {}
    a_failed = False
    b_failed = False
    for attr_index, component in enumerate(nx.connected_components(dep_graph)):
        comp_subgraph = dep_graph.subgraph(component)
        a_number = sum(
            1
            for node, keys in comp_subgraph.nodes(data=True)
            if keys['base'] == 'A'
        )
        b_number = sum(
            1
            for node, keys in comp_subgraph.nodes(data=True)
            if keys['base'] == 'B'
        )

        if a_number == 0 and bijective:
            # Some attrs of target graph may be missing in source if bijection is not required
            a_failed = True
        if b_number == 0:
            # This condition means that there are no appropriate nodes/edges in B for node/edge of A to be mapped to
            b_failed = True

        num_edges = comp_subgraph.number_of_edges()
        if num_edges != a_number * b_number:
            raise ApplicabilityScopeError(f"Unable to create valid {item_type} attributes for {repr(match_function)}")

        for node, keys in comp_subgraph.nodes(data=True):
            if keys['base'] == 'A':
                a_attrs[keys['name']] = attr_index
            else:
                b_attrs[keys['name']] = attr_index

    if a_failed:
        a_attrs = None
    if b_failed:
        b_attrs = None
    return a_attrs, b_attrs


def _get_log_function(verbose: bool) -> Callable[[str], None]:
    if verbose:
        return lambda message: print(f"[VF3Py] {message}")
    else:
        return lambda message: None


@_format_docstring
def main_vf3_caller(
        subgraph: GRAPH_TYPE,
        graph:    GRAPH_TYPE,
        node_match: Optional[Callable[[dict, dict], bool]] = None,
        edge_match: Optional[Callable[[dict, dict], bool]] = None,
        all_solutions: bool=True,
        return_integers: bool=False,
        verbose: bool=False,
        variant: VF3_VARIANTS_TYPE = 'B',
        num_threads: int=1,
        induced: bool=True,
        bijective_for_nodes: bool=False,
    ) -> List[Dict]:
    """The core routine of the VF3Py library. It solves the subgraph iso/monomorphism problem, i.e. finds ways to match nodes of ``subgraph`` with some/all nodes of ``graph``. Only NetworkX graphs are accepted (either nx.Graph or nx.DiGraph). **NOTE**: It is not practical to call this function directly -- use one of the front-end functions instead: ``are_isomorphic``, ``get_exact_isomorphisms``, ``get_automorphisms``, ``has_subgraph``, ``get_subgraph_isomorphisms``.

    Args:
        {SUBGRAPH}
        {LARGE_GRAPH}
        {NODE_MATCH}
        {EDGE_MATCH}
        {ALL_SOLUTIONS}
        {RETURN_INTEGERS}
        {VARIANT}
        {NUM_THREADS}
        {VERBOSE}
        {INDUCED}
        {BIJECTIVE_FOR_NODES}

    Returns:
        List[Dict]: List of subgraph->graph isomorphisms. Each isomorphism is represented by a dict that maps: 'subgraph' labels -> 'graph' labels.
    """
    
    log = _get_log_function(verbose)

    if bijective_for_nodes and (subgraph.number_of_nodes() != graph.number_of_nodes()):
        log("Initial check indicates the absence of solutions")
        return []
    
    assert num_threads == 1 or variant == 'P', "Only VF3P variant supports multithreading"
    assert num_threads >= 1 and num_threads <= MAX_THREADS, f"Number of threads must be between 1 and {MAX_THREADS}"
    assert variant in VF3_VARIANTS, f"VF3 variant must be one of {VF3_VARIANTS}"
    if variant == 'P':
        assert num_threads > 1, "Parallel version should use more than one thread. Set num_threads accordingly"

    _ensure_graph_correctness(graph)
    _ensure_graph_correctness(subgraph)
    log("Graph correctness checks were passed")

    target_dict, target_labels, target_directed = _process_graph(graph)
    pattern_dict, pattern_labels, pattern_directed = _process_graph(subgraph)
    assert not (target_directed ^ pattern_directed), \
        f"Both graphs must be either directed or undirected"
    directed = target_directed
    log("Graphs were loaded successfully")

    if node_match is not None:
        log("Generating node attributes")
        subgraph_node_attrs, graph_node_attrs = _group_attrs(subgraph.nodes, graph.nodes, node_match, 'node',
                                                             bijective=bijective_for_nodes)
        if graph_node_attrs is None or subgraph_node_attrs is None:
            log("Initial check indicates the absence of solutions")
            return []
        
        use_node_attrs = (len(set(graph_node_attrs.values())) > 1) and (len(set(subgraph_node_attrs.values())) > 1)
        if use_node_attrs:
            target_dict['node_attrs'] = [
                graph_node_attrs[original_label]
                for original_label in target_labels['from_int']
            ]
            pattern_dict['node_attrs'] = [
                subgraph_node_attrs[original_label]
                for original_label in pattern_labels['from_int']
            ]
            log("Node attributes were generated successfully")
        else:
            log("Use of node attributes is redundant")
    else:
        log("Skipping node attributes generation")
        use_node_attrs = False

    if edge_match is not None:
        log("Generating edge attributes")
        subgraph_edge_attrs, graph_edge_attrs = _group_attrs(subgraph.edges, graph.edges, edge_match, 'edge',
                                                             bijective=(bijective_for_nodes and induced))
        if graph_edge_attrs is None or subgraph_edge_attrs is None:
            log("Initial check indicates the absence of solutions")
            return []
        
        use_edge_attrs = (len(set(graph_edge_attrs.values())) > 1) and (len(set(subgraph_edge_attrs.values())) > 1)
        log(f"graph_edge_attrs = {repr(graph_edge_attrs)}")
        log(f"subgraph_edge_attrs = {repr(subgraph_edge_attrs)}")
        if use_edge_attrs:
            target_dict['edge_attrs'] = [
                graph_edge_attrs[target_labels['from_int'][vA], target_labels['from_int'][vB]]
                for vA, vB in target_dict['edges']
            ]
            pattern_dict['edge_attrs'] = [
                subgraph_edge_attrs[pattern_labels['from_int'][vA], pattern_labels['from_int'][vB]]
                for vA, vB in pattern_dict['edges']
            ]
            log("Edge attributes were generated successfully")
        else:
            log("Use of edge attributes is redundant")
    else:
        log("Skipping edge attributes generation")
        use_edge_attrs = False

    log("Loading finished. Entering C++ part...")
    result = VF3_CALLS[variant, use_node_attrs, use_edge_attrs](
        target=target_dict, pattern=pattern_dict,
        directed=directed, all_solutions=all_solutions, verbose=verbose, induced=induced,
        num_threads=num_threads,
    )
    log(f"Returned to Python. Found {len(result)} solutions")
    if not return_integers:
        result = [
            {
                pattern_labels['from_int'][source]: target_labels['from_int'][target]
                for source, target in match_data
            }
            for match_data in result
        ]
        log("Successfully translated solutions to the original node labels")
    return result


# FRONT-END FUNCTIONS
# Exact matches

@_format_docstring
def are_isomorphic(source_graph: GRAPH_TYPE,
                   target_graph: GRAPH_TYPE,
                   get_mapping: bool = False,
                   **kwargs) -> Union[bool, Tuple[bool, Union[Dict, None]]]:
    """Check if two graphs are isomorphic. This includes checks for number of nodes - ``are_isomorphic`` always returns False if two graphs have different number of nodes, as opposed to graph-subgraph isomorphisms.

    Args:
        {FIRST_GRAPH}
        {SECOND_GRAPH} Swap between source_graph <-> target_graph changes nothing.
        {NODE_MATCH}
        {EDGE_MATCH}
        {VARIANT}
        {NUM_THREADS}
        {VERBOSE}

    Returns:
        bool: True if graphs are isomorphic, False - otherwise.
        Also, optionally returns Dict | None. The Dict represents the mapping 'target_graph' labels -> 'source_graph' labels. Returned only if get_mapping set to True.
    """

    isom_list = main_vf3_caller(source_graph,
                                target_graph,
                                bijective_for_nodes=True,
                                all_solutions=False,
                                **kwargs)
    if get_mapping:
        return len(isom_list) > 0, isom_list[0] if len(isom_list) > 0 else None
    else:
        return len(isom_list) > 0


@_format_docstring
def get_exact_isomorphisms(source_graph: GRAPH_TYPE, target_graph: GRAPH_TYPE, **kwargs) -> List[Dict]:
    """Get a list of all isomorphisms between two NetworkX graphs. This includes checks for number of nodes - ``get_exact_isomorphisms`` always returns ``[]`` if two graphs have different number of nodes, as opposed to graph-subgraph isomorphisms.

    Args:
        {FIRST_GRAPH}
        {SECOND_GRAPH} Swap between source_graph <-> target_graph swaps keys with values in the resulting isomorphisms dicts.
        {NODE_MATCH}
        {EDGE_MATCH}
        {RETURN_INTEGERS}
        {VARIANT}
        {NUM_THREADS}
        {VERBOSE}

    Returns:
        List[Dict]: List of source_graph<->target_graph isomorphisms. Each isomorphism is represented by a dict that maps: 'target_graph' labels -> 'source_graph' labels.
    """
    return main_vf3_caller(source_graph, target_graph,
                           bijective_for_nodes=True,
                           all_solutions=True,
                           **kwargs)


@_format_docstring
def get_automorphisms(graph: GRAPH_TYPE, **kwargs) -> List[Dict]:
    """Get isomorphic mappings of NetworkX graph onto itself (automorphism).

    Args:
        {ANY_GRAPH}
        {NODE_MATCH}
        {EDGE_MATCH}
        {RETURN_INTEGERS}
        {VARIANT}
        {NUM_THREADS}
        {VERBOSE}

    Returns:
        List[Dict]: List of graph<->graph isomorphisms (represented as dicts).
    """
    return main_vf3_caller(graph, graph,
                           bijective_for_nodes=True,
                           all_solutions=True,
                           **kwargs)


# FRONT-END FUNCTIONS
# Subgraph matches

@_format_docstring
def has_subgraph(
        subgraph: GRAPH_TYPE, graph: GRAPH_TYPE, get_mapping: bool = False, **kwargs
    ) -> Union[bool, Tuple[bool, Union[Dict, None]]]:
    """Check if `subgraph` is a subgraph of `graph` using VF3 algorithm.

    Args:
        {SUBGRAPH}
        {LARGE_GRAPH}
        {NODE_MATCH}
        {EDGE_MATCH}
        {RETURN_INTEGERS}
        {VARIANT}
        {NUM_THREADS}
        {VERBOSE}

    Returns:
        bool: True if `subgraph` is, indeed, a subgraph of `graph`, False - otherwise.
        Also, optionally returns Dict | None: subgraph->graph isomorphism represented by a dict that maps: 'subgraph' labels -> 'graph' labels. Returned only if get_mapping set to True.
    """
    isom_list = main_vf3_caller(subgraph, graph, all_solutions=False, **kwargs)
    if get_mapping:
        return len(isom_list) > 0, isom_list[0] if len(isom_list) > 0 else None
    else:
        return len(isom_list) > 0


@_format_docstring
def get_subgraph_isomorphisms(subgraph: GRAPH_TYPE, graph: GRAPH_TYPE, **kwargs) -> List[Dict]:
    """Solve subgraph isomorphism problem, i.e. find ways to match nodes of `subgraph` with some/all nodes of `graph`.

    Args:
        {SUBGRAPH}
        {LARGE_GRAPH}
        {NODE_MATCH}
        {EDGE_MATCH}
        {RETURN_INTEGERS}
        {VARIANT}
        {NUM_THREADS}
        {VERBOSE}

    Returns:
        List[Dict]: List of subgraph->graph isomorphisms. Each isomorphism is represented by a dict that maps: 'subgraph' labels -> 'graph' labels.
    """
    return main_vf3_caller(subgraph, graph, all_solutions=True, **kwargs)

@_format_docstring
def has_monomorphic_subgraph(
        subgraph: GRAPH_TYPE, graph: GRAPH_TYPE, get_mapping: bool = False, **kwargs
    ) -> Union[bool, Tuple[bool, Union[Dict, None]]]:
    """Check if `subgraph` is a subgraph of `graph` using VF3 algorithm.

    Args:
        {SUBGRAPH}
        {LARGE_GRAPH}
        {NODE_MATCH}
        {EDGE_MATCH}
        {RETURN_INTEGERS}
        {VARIANT}
        {NUM_THREADS}
        {VERBOSE}

    Returns:
        bool: True if `subgraph` is, indeed, a subgraph of `graph`, False - otherwise.
        Also, optionally returns Dict | None: subgraph->graph isomorphism represented by a dict that maps: 'subgraph' labels -> 'graph' labels. Returned only if get_mapping set to True.
    """

    isom_list = main_vf3_caller(subgraph, graph, induced=False, all_solutions=False, **kwargs)
    if get_mapping:
        return len(isom_list) > 0, isom_list[0] if len(isom_list) > 0 else None
    else:
        return len(isom_list) > 0

@_format_docstring
def get_subgraph_monomorphisms(subgraph: GRAPH_TYPE, graph: GRAPH_TYPE, **kwargs) -> List[Dict]:
    """Solve subgraph monomorphism problem, i.e. find ways to match nodes of `subgraph` with some/all nodes of `graph`.

    Args:
        {SUBGRAPH}
        {LARGE_GRAPH}
        {NODE_MATCH}
        {EDGE_MATCH}
        {RETURN_INTEGERS}
        {VARIANT}
        {NUM_THREADS}
        {VERBOSE}

    Returns:
        List[Dict]: List of subgraph->graph isomorphisms. Each isomorphism is represented by a dict that maps: 'subgraph' labels -> 'graph' labels.
    """
    return main_vf3_caller(subgraph, graph, induced=False, all_solutions=True, **kwargs)



import glob
import ntpath
import sys
import os

import vf3py

import networkx as nx
from networkx.algorithms import isomorphism


SDF_TEMPLATE = os.path.join(os.path.dirname(__file__), './mols/*.sdf')
MOL_FILES = {
    ntpath.basename(file.replace('.sdf', '')): file
    for file in glob.glob(SDF_TEMPLATE)
}


def sdf_to_graph(sdf_fname: str) -> nx.Graph:
    molgraph = nx.Graph()

    with open(sdf_fname, "r") as f:
        lines = f.readlines()
    natoms = int(lines[3][0:3])
    nbonds = int(lines[3][3:6])
    for i in range(4, 4 + natoms):
        molgraph.add_node(i-4)
        parts = lines[i].replace("\n", "").split()
        molgraph.nodes[i-4]['symbol'] = parts[3]
    for i in range(4 + natoms, 4 + natoms + nbonds):
        at1 = int(lines[i][0:3])
        at2 = int(lines[i][3:7])
        bondtype = int(lines[i][7:10])
        molgraph.add_edge(at1 - 1, at2 - 1)
        molgraph[at1 - 1][at2 - 1]['type'] = bondtype
    return molgraph


def same_element(n1_attrib: dict, n2_attrib: dict) -> bool:
    return n1_attrib['symbol'] == n2_attrib['symbol']


def get_hcarbon_subgraph(graph) -> nx.Graph:
    save_atoms = []
    for node in graph.nodes:
        if graph.nodes[node]['symbol'] == 'H':
            nb_list = list(graph.neighbors(node))
            if len(nb_list) == 1 and graph.nodes[nb_list[0]]['symbol'] == 'C':
                continue
        save_atoms.append(node)
    subgraph = graph.subgraph(save_atoms)

    return subgraph


def generate_isomorphisms(graph):
    GM = isomorphism.GraphMatcher(graph, graph, node_match=same_element)
    isomorphisms = set()
    atoms = list(graph.nodes)
    for isom in GM.isomorphisms_iter():
        isomorphisms.add(tuple(isom[i] for i in atoms))
    return isomorphisms


def main(verbose=False):
    for molname, molfile in MOL_FILES.items():
        # print(f"Processing test molecule '{molname}'")
        mol_graph = sdf_to_graph(molfile)
        subgraph = get_hcarbon_subgraph(mol_graph)
        reference_answer = generate_isomorphisms(subgraph)

        atoms = list(subgraph.nodes)
        for variant in ('B', 'P', 'L'):
            result = set(
                tuple(
                    isom[i]
                    for i in atoms
                )
                for isom in vf3py.get_automorphisms(subgraph, node_match=same_element,
                                                    verbose=verbose, variant=variant,
                                                    num_threads=vf3py.MAX_THREADS if variant == 'P' else 1)
            )
            assert result == reference_answer, \
                f"Failed for test graph '{molname}': correct={repr(reference_answer)},\ngot={repr(result)}"
    print("Molecular isomorphisms test passed")


if __name__ == "__main__":
    VERBOSE = '-v' in sys.argv
    main(verbose=VERBOSE)



import sys
import networkx as nx
from networkx.algorithms import isomorphism

import vf3py


def same_edge(n1_attrib: dict, n2_attrib: dict) -> bool:
    return n1_attrib['type'] == n2_attrib['type']


def main(verbose=False):
    graph = nx.DiGraph()
    graph.add_edges_from([
        ('A',  'C', {'type': 1}),
        ('A',  'B', {'type': 1}),
        ('A',  'B`',{'type': 1}),
        ('B',  'C', {'type': 1}),
        ('B`', 'C', {'type': 2}),
    ])

    if verbose:
        for edge in graph.edges(data=True):
            print(repr(edge))

    GM = isomorphism.GraphMatcher(graph, graph, edge_match=same_edge)
    reference_answer = set()
    atoms = list(graph.nodes)
    for isom in GM.isomorphisms_iter():
        reference_answer.add(tuple(isom[i] for i in atoms))

    if verbose:
        print(f"REF = {repr(reference_answer)}")
    result = set(
        tuple(
            isom[i]
            for i in atoms
        )
        for isom in vf3py.get_automorphisms(graph, edge_match=same_edge, verbose=verbose)
    )
    assert result == reference_answer, \
        f"Failed for test graph : correct={repr(reference_answer)},\ngot={repr(result)}"
    print('Edge test passed')


if __name__ == "__main__":
    VERBOSE = '-v' in sys.argv
    main(verbose=VERBOSE)



def run_tests():
    from .basic_graphs import main as f_basic_graphs
    from .edge_test import main as f_edge_attrs
    from .mol_isoms import main as f_molecular
    from .mono import main as f_mono
    from .mono_random import main as f_mono_random
    from .simple_attrs import main as f_simple_attrs

    TEST_FUNCTIONS = {
        'basic_graphs': f_basic_graphs,
        'edge_attrs': f_edge_attrs,
        'molecular': f_molecular,
        'monomorphism': f_mono,
        'random monomorphism': f_mono_random,
        "Anton's example": f_simple_attrs,
    }

    for function_name, function in TEST_FUNCTIONS.items():
        print(f"=== Running test '{function_name}' ===")
        function()
    
    print("OK")


if __name__ == "__main__":
    run_tests()



from ..vf3py_base import calc_noattrs, calc_nodeattr, calc_edgeattr, calc_bothattrs
from ..vf3py_vf3l import calc_l_noattrs, calc_l_nodeattr, calc_l_edgeattr, calc_l_bothattrs
from ..vf3py_vf3p import calc_p_noattrs, calc_p_nodeattr, calc_p_edgeattr, calc_p_bothattrs



PACKAGE_DATA = {'package': 'vf3py', 'module': 'vf3py', 'version': '0.0.10', 'author': 'Nikolai Krivoshchapov', 'platforms': ['Linux'], 'python_versions': '>=3.8.0', 'install_requires': ['numpy', 'networkx'], 'project_summary': 'Interfacing Python and NetworkX with VF3 – the fastest algorithm for graph/subgraph isomorphism calculation', 'project_urls': 'Project-URL: Gitlab home page, https://gitlab.com/knvvv/vf3py\nProject-URL: Docs, https://knvvv.gitlab.io/vf3py', 'readme_content': 'Description-Content-Type: text/markdown\n\n# VF3Py\n\nInterfacing Python and NetworkX with VF3 – the fastest algorithm for graph/subgraph isomorphism calculation.\n\n## Setup\n\nMake sure that your OS is Linux and Python version is >= 3.8.\n\nVF3Py can be installed using this command:\n\n```\npip install vf3py\n```\nTest your installation:\n\n```\n$ python\n>>> import vf3py.test\n>>> vf3py.test.run_tests()\n(...lots of output...)\nOK\n>>> \n```\n\n## Documentation\n\n[![pipeline status](https://gitlab.com/knvvv/vf3py/badges/master/pipeline.svg)](https://gitlab.com/knvvv/vf3py/-/commits/master)\n\nAvailable [here](https://knvvv.gitlab.io/vf3py)\n\n\n## Other projects that use VF3Py\n\nFor now, all projects are mine:\n\n* [algebra_repr](https://gitlab.com/knvvv/algebra_repr) (TODO: come up with a better name) -- uses SageMath to construct finite algebraic structures (groups, rings, fields), then represents them as graphs. Graph isomorphism allows to check if any two algebraic structures are isomorphic or not.\n\n* [PyXYZ](https://gitlab.com/knvvv/pyxyz) -- A Python Library for Molecular Geometry Manipulation. TODO: Use VF3Py to generate automorphisms groups for molecular graphs.\n\n\n## Links\n\n[Gitlab home page](https://gitlab.com/knvvv/vf3py)\n\n[VF3Py PyPi page](https://pypi.org/project/vf3py)\n\n[Original VF3 implementation](https://github.com/MiviaLab/vf3lib)\n\n\n## References\n\n1. Challenging the time complexity of exact subgraph isomorphism for huge and dense graphs with VF3 - Carletti V., Foggia P., Saggese A., Vento M. - IEEE transactions on pattern analysis and machine intelligence - 2018\n2. Introducing VF3: A new algorithm for subgraph isomorphism - Carletti V., Foggia P., Saggese A., Vento M. - International Workshop on Graph-Based Representations in Pattern Recognition - 2017\n3. Comparing performance of graph matching algorithms on huge graphs - Carletti V., Foggia P., Saggese A., Vento M. - Pattern Recognition Letters - 2018\n4. A Parallel Algorithm for Subgraph Isomorphism - V. Carletti, P. Foggia, P. Ritrovato, M. Vento, V. Vigilante - International Workshop on Graph-Based Representations in Pattern Recognition - 2019\n', 'files': '', 'install_requires_lines': 'numpy\nnetworkx', 'platforms_lines': 'Platform: Linux'}
import platform
import glob
import ntpath
import os
import json
import shutil
from setuptools import setup

# OS specifics
CUR_OS = platform.system()
SHAREDOBJ_TEMPLATE = {
    'Windows': ["vf3py_base.cp{py_ver}-win_amd64.pyd", "vf3py_vf3l.cp{py_ver}-win_amd64.pyd", "vf3py_vf3p.cp{py_ver}-win_amd64.pyd"],
    'Linux': ["vf3py_base.cpython-{py_ver}*-x86_64-linux-gnu.so", "vf3py_vf3l.cpython-{py_ver}*-x86_64-linux-gnu.so", "vf3py_vf3p.cpython-{py_ver}*-x86_64-linux-gnu.so"],
}

assert CUR_OS in ['Linux', 'Windows'], "Only Linux and Windows platforms are supported"

if CUR_OS == 'Windows':
    DLLDEPS_JSON = 'win_dlldeps.json'
    DLL_STORAGE_DIR = 'win_dlls'
    assert os.path.isfile(DLLDEPS_JSON), f'Required file "{DLLDEPS_JSON}" not found'

# Python version specifics
python_version_tuple = platform.python_version_tuple()
py_ver = int(f"{python_version_tuple[0]}{python_version_tuple[1]}")

object_names = []
for somask in SHAREDOBJ_TEMPLATE[CUR_OS]:
    so_list = glob.glob(os.path.join('./vf3py', somask.format(py_ver=py_ver)))
    assert len(so_list) == 1
    object_names.append(ntpath.basename(so_list[0]))

for file in glob.glob('./vf3py/*.pyd') + glob.glob('./vf3py/*.so'):
    if ntpath.basename(file) not in object_names:
        os.remove(file)

if CUR_OS == 'Windows':
    assert os.path.isfile(DLLDEPS_JSON), f'Required file "{DLLDEPS_JSON}" not found'

    with open(DLLDEPS_JSON, 'r') as f:
        dlldeps_data = json.load(f)
    
    for object_name in object_names:
        assert object_name in dlldeps_data, f"'{object_name}' is not accounted in {DLLDEPS_JSON}"
        for file in dlldeps_data[object_name]:
            shutil.copy2(os.path.join(DLL_STORAGE_DIR, file), './vf3py')
    
    ADDITIONAL_FILES = ['*.dll']

elif CUR_OS == 'Linux':
    ADDITIONAL_FILES = []

setup(
    name=PACKAGE_DATA['package'],
    version=PACKAGE_DATA['version'],
    author=PACKAGE_DATA['author'],
    python_requires=f'=={python_version_tuple[0]}.{python_version_tuple[1]}.*',
    install_requires=PACKAGE_DATA['install_requires'],
    platforms=PACKAGE_DATA['platforms'],
    packages=[PACKAGE_DATA['package']],
    package_data={'vf3py': ['__init__.py', *object_names, 'cpppart/*.py*', 'test/*.py', 'test/mols/*.sdf', *ADDITIONAL_FILES]}
)



import sys
import networkx as nx
from networkx.algorithms import isomorphism

import vf3py

def main(verbose=False):
    nx_graph = nx.Graph()
    
    nx_graph.add_edge(1, 2, color="red")
    nx_graph.add_edge(2, 3, color="blue")
    nx_graph.add_edge(3, 4, color="green")
    nx_graph.add_edge(4, 1, color="yellow")
    nx_graph.add_edge(4, 2, color="blue")
    nx_graph.add_edge(3, 1, color="yellow")

    nx_graph.add_nodes_from([3, 4], color='red')
    nx_graph.add_nodes_from([1], color='blue')
    nx_graph.add_nodes_from([2], color='green')
 
    nx_graph_sub = nx.Graph()
    nx_graph_sub.add_edge('A', 'B', color="red")
    nx_graph_sub.add_nodes_from('A', color='blue')
    nx_graph_sub.add_nodes_from('B', color='green')

    def node_match(subgraph_dict, graph_dict):
        return graph_dict['color'] == subgraph_dict['color']
    
    result = vf3py.get_subgraph_isomorphisms(
        subgraph=nx_graph_sub,
        graph=nx_graph,
        node_match=node_match
    )
    assert result == [{'A': 1, 'B': 2}]


if __name__ == "__main__":
    VERBOSE = '-v' in sys.argv
    main(verbose=VERBOSE)



import vf3py
import networkx as nx
from networkx.algorithms.isomorphism import GraphMatcher
import random

def nx_is_monomorphic(subg, g):
    gm = GraphMatcher(g, subg)
    return gm.subgraph_is_monomorphic()

# Hardcoded sanity check test case.
def main():
    for num_nodes in range(10, 20, 5):
        for ita in range(2, 5):
            num_edges = int(num_nodes * num_nodes * ita / 10.0)
            print(f"Testing with {num_nodes} nodes and {num_edges} edges.")

            g1 = nx.gnm_random_graph(num_nodes, num_edges)
            g2 = nx.Graph()

            for e in nx.edges(g1):
                if random.random() > 0.5:
                    g2.add_edge(e[0], e[1])
            
            print(f"G1: {len(g1.nodes())}-{len(g1.edges())}, G2: {len(g2.nodes())}-{len(g2.edges())}")
            for variant in ('B', 'P', 'L'):
                print(f"Running VF3{variant}...")
                if not vf3py.has_monomorphic_subgraph(g2, g1, variant=variant,
                                                      num_threads=vf3py.MAX_THREADS if variant == 'P' else 1):
                    # VF3 Wrong?
                    print("Running nx.VF2...")
                    assert not nx_is_monomorphic(g2, g1)
                    

    print("Random monomorphism test complete.")

if __name__ == "__main__":
    main()